#!/usr/bin/env python

############################################################################
#    Copyright (C) 2004 by                                                 #
#    Tero Kuusela (teroajk at users dot sourceforge dot net)               #
#    Stefan Brockmann (stefankjb at yahoo dot com)                         #
#    Petteri Klemola (pzq at users dot sourceforge dot net)                #
#                                                                          #
#    This program is free software; you can redistribute it and/or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

import pygame
import os
import math
import penguin
import snowball
import highscore
from gfx_functions import *
from pygame.locals import *


# python 2.2 does not have math.radians(degrees) function
# use this function instead
def deg2rad(degrees):
    return degrees * math.pi / 180.0

# Our event name.
# To be used heavily when support for multiple events is added.
EVENT_NAME = 'snowballthrow'

# These are the phases the game can be in.
PHASE_RUN = 0
PHASE_AIM = 1
PHASE_THROW = 2
PHASE_FLIGHT = 3
PHASE_RESULT = 4
PHASE_FOUL = 5
PHASE_FUMBLE = 6
PHASE_HELP = 7

# These variables are used to control the speed-gaining algorithm
MAX_SPEED = 200          # maximum running speed befor falling down
SPEED_LOSS = 1           # how much speed we lose at each tick
MAX_SPEED_GAIN = 55      # how much speed we can gain at one time
OPTIMAL_AT_START = 200   # the optimal delay at the beginning
OPTIMAL_CHANGE = 0.92    # the rate of change of optimal delay

# Controls how many pixels we move the penguin at each step
# The actual movement is self.speed / RUN_ANIM_SPEED
RUN_ANIM_SPEED = 35

STAND_X, STAND_Y = 20.0, 160.0 # Penguin position in screen 2   # 40, 320
ICY_LINE = 190 # Yliastumisraja  # 380
FLIGHT_SPEED_FACTOR = 4.0

# These variables control the sliding animation
SLIDE_START_SPEED = 0.8   # How fast we begin to slide
SLIDE_ACCELERATION = 1.1  # How quickly the slide accelerates

# Color values used frequently.
COLOR_BGFILL = (220, 230, 255) # 220, 230, 255
COLOR_TEXT = (21,65,125)   # 42, 131, 251

class SnowBallThrowGame:
    "Snow ball throwing game for Tuxletics."

    def __init__(self, screen,players): #, players, highscore):

        # We begin in running phase.
        self.phase = PHASE_RUN

        self.clock = pygame.time.Clock()
        self.screen = screen
        self.players = players

        fontdir = os.getcwd()
        self.font = pygame.font.Font(os.path.join(fontdir, "Domestic_Manners.ttf"), 15)  # 30
        self.s_font = pygame.font.Font(os.path.join(fontdir, "Vera.ttf"), 8)  # 16

        # This is our penguin contestant.
        self.penguin = penguin.Penguin()
        self.p_sprite = pygame.sprite.RenderPlain(self.penguin)

        # This is the snowball we throw.
        self.ball = snowball.Snowball()
        self.ball_sprite = pygame.sprite.RenderPlain(self.ball)

        # Here we set up the screen for this event.
        pygame.display.set_caption('Tuxletics - Snowball throwing')
        self.bg = pygame.Surface(screen.get_size()).convert()
        self.screen1_img, self.screen1_rect = load_png("snowballthrow/screen1.png")
        self.screen2_img, self.screen2_rect = load_png("snowballthrow/screen2.png")
        self.img, self.rect = self.screen1_img, self.screen1_rect

        # These load and place the indicators for throwing angle and speed.
        self.aim_image, self.aim_rect = load_png("snowballthrow/aim.png")
        self.aim_rect.move_ip(250, 10)  # 500, 20

        self.a_text = self.font.render("Angle", 1, COLOR_TEXT)
        self.a_rect = self.a_text.get_rect()
        self.a_rect.move_ip(275, 2) # 550, 5

        self.speed_image, self.speed_rect = load_png("snowballthrow/aim.png")
        self.speed_image = pygame.transform.flip(self.speed_image, 0, 1)
        self.speed_rect.move_ip(250, 60)  # 500, 120

        self.s_text = self.font.render("Speed", 1, COLOR_TEXT)
        self.s_rect = self.s_text.get_rect()
        self.s_rect.move_ip(275, 92)  # 550, 185

        ptxt = "Player: "+self.players[0].name
        self.p_text = self.font.render(ptxt, 1, COLOR_TEXT)
        self.p_rect = self.p_text.get_rect()
        self.p_rect.move_ip(5, 0)   # 10

        #self.players=players
        #self.highscore=highscore

        self.start_event()

    def start_event(self):
        print "EVENT: SNOW BALL THROWING" # DEBUG

        # Old keypresses can make us do silly things.
        pygame.event.clear()

        # We begin our penguin animation from the beginning...
        self.penguin.reinit()
        self.penguin.set_visible_image(0)

        # ...and redraw the screen with the starting position.
        self.bg.blit(self.img, self.rect)
        self.screen.blit(self.bg, (0, 0))
        self.screen.blit(self.penguin.image, self.penguin.rect)
        pygame.display.flip()

        # Reinit necessary variables.
        self.speed=0            # stores the current running speed
        self.foot1_changed = 0  # used to track key presses while running
        self.foot2_changed = 0  # two are needed to track the keypress order
        self.travelled = 0      # travelled distance when running, for animation
        self.angle=0            # throwing angle of the snowball
        self.flight_len = 0     # length of our throw
        self.throw_behind_line = 0 # how far behind line ball is thrown

        self.optimal = OPTIMAL_AT_START  # optimal delay between keypresses
        self.interval = 0                # used when calculating speed gain

        self.phase = PHASE_RUN
        self.quit = 0

    def play(self):
        while not self.quit:
            if self.phase == PHASE_RUN:      self.gain_speed()
            elif self.phase == PHASE_AIM:    self.aim_angle()
            elif self.phase == PHASE_THROW:  self.throw_ball()
            elif self.phase == PHASE_FLIGHT: self.ball_flight()
            elif self.phase == PHASE_RESULT: self.show_result()
            elif self.phase == PHASE_FOUL:   self.foul()
            elif self.phase == PHASE_FUMBLE: self.fumble()
            elif self.phase == PHASE_HELP:   self.show_help()
            else:
                print "Unk phase" # DEBUG
                self.quit = 1

    def gain_speed(self):
        print "GAINING SPEED..." # DEBUG

        # change background
        self.img, self.rect = self.screen1_img, self.screen1_rect
        self.bg.blit(self.img, self.rect)

        self.render()
        self.show_text("Press L for help", (65, 230))  # 130, 460

        # These are used to calculate the speed gain.
        foot1 = pygame.time.get_ticks()
        foot2 = pygame.time.get_ticks()

        while True:
            # Restrict fps to 60
            self.clock.tick(60)

            if self.speed > 0:
                # We are running.
                self.speed -= SPEED_LOSS
                if self.speed > RUN_ANIM_SPEED:  # To keep animation reasonable
                    # Animate the running phase, using the running speed to
                    # calculate the speed of animation.
                    self.penguin_run()

                # We may not step on the icy area.
                if self.penguin.rect.left >= ICY_LINE:
                    self.phase = PHASE_FOUL
                    return
                self.render()

            for event in pygame.event.get():
                if self.quit_event(event):
                    return
                elif event.type == pygame.KEYDOWN:
                    # We have two control buttons that need to be pressed
                    # alternately to gain speed.
                    if event.key == K_LEFT:
                        foot1 = pygame.time.get_ticks()
                        self.foot1_changed = 1
                    elif event.key == K_RIGHT:
                        foot2 = pygame.time.get_ticks()
                        self.foot2_changed = 1
                    elif event.key == K_SPACE:
                        self.foot1_changed = 0
                        self.foot2_changed = 0
                        self.throw_behind_line = ICY_LINE - self.penguin.rect.left # how far behind line ball is thrown
                        self.throw_behind_line *= 0.1 # scale is bigger in screen 1
                        self.phase = PHASE_AIM
                        return
                    elif event.key == K_TAB:
                        self.phase = PHASE_HELP
                        return
                # If the keys have been pressed alternately, we calculate the
                # speed increase from the time between the keypresses.
                if (self.foot1_changed and self.foot2_changed):
                    self.calculate_speed_gain(foot1, foot2)
                    # If we go too fast, we fall down
                    if self.speed > MAX_SPEED:
                        self.phase = PHASE_FUMBLE
                        return
        print "We should never reach this!"
        sys.exit(1)

    def calculate_speed_gain(self, foot1, foot2):
        # What was the interval between keypresses?
        self.interval = math.fabs(foot2 - foot1)

        # The closer we get to optimal, the more speed we gain.
        if self.interval > self.optimal:
            t = self.optimal / self.interval
        else:
            t = self.interval / self.optimal

        # We gain t% of MAX_SPEED_GAIN speed
        self.speed += t * MAX_SPEED_GAIN
        # Shorten the optimal interval
        self.optimal *= OPTIMAL_CHANGE
        if self.optimal < 20:
            self.optimal = 40

        self.foot1_changed = 0
        self.foot2_changed = 0

    def aim_angle(self):
        self.angle = 0.0

        while self.phase==PHASE_AIM:
            # Restrict fps to 60
            self.clock.tick(60)

            for event in pygame.event.get():
                if self.quit_event(event):
                    return
                elif event.type == pygame.KEYUP: # key released..
                    if event.key == K_SPACE:
                        self.phase = PHASE_THROW
                        return

            # The more speed we have, the harder it is to set the angle
            self.angle += 2.5 * (0.5 + 0.5 * float(self.speed) / float(MAX_SPEED))
            if self.angle>90.0:
                self.angle=90.0
                self.phase = PHASE_THROW

            self.penguin.set_visible_image( 7 )
            self.render()

    def ball_flight(self):
        ground_level = 182     # 365
        self.flight_len = 0
        x, y = STAND_X - self.throw_behind_line + 7, STAND_Y+10.0   # 15, 20

        # 0.5 of ball velocity from throwing and 0...0.5 from speeding
        ball_velocity = 0.5 + 0.5 * float(self.speed) / float(MAX_SPEED)
        ball_velocity *= FLIGHT_SPEED_FACTOR

        dx = math.cos(deg2rad(self.angle)) * ball_velocity
        dy = -math.sin(deg2rad(self.angle)) * ball_velocity

        # change background
        self.img, self.rect = self.screen2_img, self.screen2_rect
        self.bg.blit(self.img, self.rect)

        while self.phase==PHASE_FLIGHT:
            # Restrict fps to 60
            self.clock.tick(60)

            for event in pygame.event.get():
                if self.quit_event(event):
                    return

            x += dx
            y += dy
            self.flight_len = x - STAND_X - self.throw_behind_line
            if self.flight_len<0:
                self.flight_len = 0
            if y >= ground_level:
                dx, dy = 0, 0
                self.phase = PHASE_RESULT
            else:
                dy += 0.002 * FLIGHT_SPEED_FACTOR * FLIGHT_SPEED_FACTOR # gravity

            self.penguin.set_visible_image( 13 )
            self.penguin.set_position(STAND_X - self.throw_behind_line, STAND_Y)
            self.ball.update(int(x), int(y), ground_level)
            self.render()

    def throw_ball(self):
        tick1 = pygame.time.get_ticks()

        while self.phase==PHASE_THROW:
            for event in pygame.event.get():
                if self.quit_event(event):
                    return

            tick2 = pygame.time.get_ticks()
            frame = 7 + 10 * (tick2-tick1) / 1000 # 10 fps

            if frame > 11:
                self.phase=PHASE_FLIGHT
                return
            self.penguin.set_visible_image( frame )
            self.render()


    def show_result(self):
        self.render()

        #Higscore is here
        hs = highscore.HighScore()

        #write score to file and get placing for player
        placing = hs.writeScoreToFile(self.players[0].name,int(self.flight_len))

        #draw scores to screen

        #first read scores from file
        scores = hs.readScoreFromFile()

        #write HighScore to screen
        line_y = 10  # 20
        self.font.set_underline(1)
        headline = self.font.render("Highscores",1,COLOR_TEXT)
        headline_rect = headline.get_rect()
        headline_rect.center = (self.rect.centerx, line_y)
        self.screen.blit(headline,headline_rect)
        self.font.set_underline(0)
        line_y += 12  # 25
        x = 0

        #write ten scores to screen highlighting players score if she made it to the list
        while x < 10:#we assume that there are 10 listings in scores-file
            if x < 9:
                tx = " "+str(x+1) +"."+" "+ scores[x].name+" "+str(scores[x].score)
            else:
                tx = str(x+1) +"."+" "+ scores[x].name+" "+str(scores[x].score)
            if placing +1 == x+1:
                txt = self.font.render(tx,1,(5,5,5))   # 10, 10, 10
            else:
                txt = self.font.render(tx,1,COLOR_TEXT)
            txt_rect = txt.get_rect()
            txt_rect.center = (self.rect.centerx, line_y)
            self.screen.blit(txt,txt_rect)
            line_y += 12   # 25
            x += 1

        #write press enter to play again to screen
        line_y += 10 # 20
        htext1 = self.font.render("Press Enter to Play again",1,(5,5,5))  # 10, 10, 10
        htext1_rect = htext1.get_rect()
        htext1_rect.center = (self.rect.centerx, line_y)
        self.screen.blit(htext1,htext1_rect)

        #write or to screen
        line_y += 20  # 40
        htext2 = self.font.render("Or",1,(5,5,5)) # 10, 10, 10
        htext2_rect = htext2.get_rect()
        htext2_rect.center = (self.rect.centerx, line_y)
        self.screen.blit(htext2,htext2_rect)

        #write press esc to go back to menu
        line_y += 20  # 40
        htext3 = self.font.render("Press Esc to go back to menu",1,(5,5,5)) # 10, 10, 10
        htext3_rect = htext3.get_rect()
        htext3_rect.center = (self.rect.centerx, line_y)
        self.screen.blit(htext3,htext3_rect)

        #update the screen
        pygame.display.flip()

        while self.phase==PHASE_RESULT:
            for event in pygame.event.get():
                if self.quit_event(event):
                    return
                elif event.type == pygame.KEYDOWN:
                    if event.key == K_RETURN:
                        self.start_event()
                        return

    def foul(self):
        print "FOUL!" # DEBUG

        p_img = 13
        i = SLIDE_START_SPEED
        while (self.penguin.rect.left < self.rect.right):
            self.p_sprite.update(i)
            # Every 5th pixel, we change the penguin image. This seems to achieve
            # pretty good animation, but could perhaps be fine tuned.
            if not self.travelled % 8:
                p_img += 1
                print "p_img ", p_img # DEBUG
                self.penguin.set_visible_image( p_img )
                if p_img is 17:
                    p_img = 15
                i *= SLIDE_ACCELERATION # Sightly accelerating slide
                self.render()
            self.travelled += 1
        self.start_event()

    def fumble(self):
        print "TOOFAAAAASSSSTTTT! *THUMP*" # DEBUG
        # Quick and dirty...
        self.p_sprite.update(2)
        self.penguin.set_visible_image( 18 )
        self.render()
        self.clock.tick(4)
        self.penguin.set_visible_image( 19 )
        # This image must be placed a lot lower than others
        self.penguin.rect.move_ip(20, 40)  # 40, 80
        self.render()
        self.clock.tick(1)
        self.start_event()

    def penguin_run(self):
        # We move the penguin one pixel forward and add one to the distance
        # we've travelled.
        self.p_sprite.update(self.speed / RUN_ANIM_SPEED)
        self.travelled += 1

        # Every 5th pixel, we change the penguin image. This seems to achieve
        # pretty good animation, but could perhaps be fine tuned.
        if not self.travelled % 5:
            self.penguin.image_id += 1
            self.penguin.set_visible_image( self.penguin.image_id )
            if self.penguin.image_id is 7:
                self.penguin.image_id = 3

    def render(self):

        # all phases render background screen and penguin
        self.screen.blit(self.bg, (0,0))
        self.p_sprite.draw(self.screen)

        # render aim and speed indicators
        if self.phase == PHASE_RUN or self.phase == PHASE_AIM or self.phase == PHASE_THROW:
            self.screen.blit(self.a_text, self.a_rect)
            self.screen.blit(self.s_text, self.s_rect)
            self.screen.blit(self.p_text, self.p_rect)
            self.screen.blit(self.aim_image, self.aim_rect)
            self.screen.blit(self.speed_image, self.speed_rect)

            # Angle indicator line.
            x = 40.0 * math.cos(deg2rad(self.angle))  # 80, 80
            y = 40.0 * math.sin(deg2rad(self.angle))
            pygame.draw.line(self.screen, (0,0,0), \
                    (self.aim_rect.left,self.aim_rect.bottom), \
                    (self.aim_rect.left+x,self.aim_rect.bottom-y), 5)

            # Keypress interval indicator line.
            xi = 40.0 * math.cos(deg2rad(self.interval / 5))   # 80, 80
            yi = 40.0 * math.sin(deg2rad(self.interval / 5))
            pygame.draw.line(self.screen, (0,0,0), \
                    (self.speed_rect.topleft), \
                    (self.speed_rect.left + xi, self.speed_rect.top + yi), 5)

            # Optimal interval indicator line.
            xo = 40.0 * math.cos(deg2rad(self.optimal / 5))   # 80, 80
            yo = 40.0 * math.sin(deg2rad(self.optimal / 5))
            pygame.draw.line(self.screen, (255,255,255), \
                    (self.speed_rect.topleft), \
                    (self.speed_rect.left + xo, self.speed_rect.top + yo), 1)

            # Current speed indicator bar.
            xs = self.speed / 2.5
            color = (self.speed + 40, 200 - self.speed, 0)   # 40, 200
            pygame.draw.line(self.screen, color, \
                    (self.speed_rect.topleft), \
                    (self.speed_rect.left + xs, self.speed_rect.top), 8)  # 8

        if self.phase == PHASE_FLIGHT or self.phase == PHASE_RESULT:

            self.screen.blit(self.p_text, self.p_rect)

            # render snowball
            self.ball_sprite.draw(self.screen)
            self.screen.blit(self.ball.shadow_image, self.ball.shadow_rect)


            # render flight length
            self.l_text = self.font.render("Length: %d" % (int(self.flight_len)),1,COLOR_TEXT)
            self.l_rect = self.l_text.get_rect()
            self.l_rect.move_ip(5, 12)  # 10, 25
            self.screen.blit(self.l_text, self.l_rect)

        if self.phase == PHASE_FUMBLE:
            self.show_text("Not so faaasssttt...", self.rect.center)

        if self.phase == PHASE_FOUL:
            self.show_text("Don't step on the ice!", self.rect.center)

        pygame.display.flip()

    def quit_event(self, event):
        if event.type == pygame.QUIT:
            self.quit = 1
            return 1
        elif event.type == pygame.KEYDOWN:
            if event.key == K_ESCAPE:
                self.quit = 1
                return 1
        return 0

    def show_help(self):
        help_file = open(os.path.join('data', EVENT_NAME, 'README'))
        help_text = help_file.readlines()
        line_y = 10  # 20
        self.screen.fill(COLOR_BGFILL)
        for line in help_text:
            txt = self.s_font.render(str(line).strip(), 1, COLOR_TEXT)
            txt_rect = txt.get_rect()
            txt_rect.center = (self.rect.centerx, line_y)
            self.screen.blit(txt, txt_rect)
            line_y += 10  # 20
        pygame.display.flip()
        while True:
            for event in pygame.event.get():
                if event.type == pygame.KEYDOWN:
                    self.phase = PHASE_RUN
                    return

    def show_text(self, text, position):
        text = self.font.render(text, 1, COLOR_TEXT)
        text_rect = text.get_rect()
        text_rect.center = position
        self.screen.blit(text, text_rect)
        pygame.display.flip()
